/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "global.h"
#include "glass_model.h"

static void _glassModelBuildVariableArray(GlassModel *model);
static void _glassModelBuildSurfaceArray(GlassModel *model);
static void _glassModelBuildActivePointArray(GlassModel *model);

/* Returns an empty model */
GlassModel *_glassModelNew(const char *name)
{
  GlassModel *model;

  model = (GlassModel *)malloc(sizeof(GlassModel));
  if(!model)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for model\n");
    return NULL;
  }
  memset(model, 0, sizeof(GlassModel));

  _glassModelSetName(model, name);

  model->variables_changed = TRUE;

  return model;
}

/* Set the model's name */
void _glassModelSetName(GlassModel *model, const char *name)
{
  if(!model)
    return;

  if(model->name)
    free(model->name);

  if(name)
    model->name = strdup(name);
  else
    model->name = strdup("");
}

/* Get the model's name */
const char *_glassModelGetName(GlassModel *model)
{
  if(!model)
    return NULL;

  return model->name;
}

/* Add a texture into the model */
void _glassModelAddTexture(GlassModel *model, GlassTexture *texture)
{
  if(!model || !texture)
    return;

  /* Append */
  if(model->last_texture)
  {
    model->last_texture->next = texture;
    model->last_texture = texture;
  }
  else
  {
    model->textures = texture;
    model->last_texture = texture;
  }
}

/* Find a texture with the name `name', return NULL if not found */
GlassTexture *_glassModelFindTexture(GlassModel *model, const char *name)
{
  GlassTexture *texture_ptr;

  if(!model || !name)
    return NULL;

  texture_ptr = model->textures;
  while(texture_ptr)
  {
    if(!strcmp(_glassTextureGetName(texture_ptr), name))
      return texture_ptr;

    texture_ptr = texture_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(model->library)
    return _glassLibraryFindTexture(model->library, name);

  return NULL;
}

/* Add a material into the model */
void _glassModelAddMaterial(GlassModel *model, GlassMaterial *material)
{
  if(!model || !material)
    return;

  /* Append */
  if(model->last_material)
  {
    model->last_material->next = material;
    model->last_material = material;
  }
  else
  {
    model->materials = material;
    model->last_material = material;
  }
}

/* Find a material with the name `name', return NULL if not found */
GlassMaterial *_glassModelFindMaterial(GlassModel *model, const char *name)
{
  GlassMaterial *material_ptr;

  if(!model || !name)
    return NULL;

  material_ptr = model->materials;
  while(material_ptr)
  {
    if(!strcmp(_glassMaterialGetName(material_ptr), name))
      return material_ptr;

    material_ptr = material_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(model->library)
    return _glassLibraryFindMaterial(model->library, name);

  return NULL;
}

/* Add a surface into the model */
void _glassModelAddSurface(GlassModel *model, GlassSurface *surface)
{
  if(!model || !surface)
    return;

  /* Append */
  if(model->last_surface)
  {
    model->last_surface->next = surface;
    model->last_surface = surface;
  }
  else
  {
    model->surfaces = surface;
    model->last_surface = surface;
  }

  /* Update the shortcut array */
  _glassModelBuildSurfaceArray(model);
}

/* Find a surface with the name `name', return NULL if not found */
GlassSurface *_glassModelFindSurface(GlassModel *model, const char *name)
{
  GlassSurface *surface_ptr;

  if(!model || !name)
    return NULL;

  surface_ptr = model->surfaces;
  while(surface_ptr)
  {
    if(!strcmp(_glassSurfaceGetName(surface_ptr), name))
      return surface_ptr;

    surface_ptr = surface_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(model->library)
    return _glassLibraryFindSurface(model->library, name);

  return NULL;
}

/* Add a variable into the model */
void _glassModelAddVariable(GlassModel *model, GlassVariable *variable)
{
  if(!model || !variable)
    return;

  /* Append */
  if(model->last_variable)
  {
    model->last_variable->next = variable;
    model->last_variable = variable;
  }
  else
  {
    model->variables = variable;
    model->last_variable = variable;
  }

  /* Update the shortcut array */
  _glassModelBuildVariableArray(model);
}

/* Find a variable with the name `name', return NULL if not found */
GlassVariable *_glassModelFindVariable(GlassModel *model, const char *name)
{
  GlassVariable *var_ptr;

  if(!model || !name)
    return NULL;

  var_ptr = model->variables;
  while(var_ptr)
  {
    if(!strcmp(_glassVariableGetName(var_ptr), name))
      return var_ptr;

    var_ptr = var_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(model->library)
    return _glassLibraryFindVariable(model->library, name);

  return NULL;
}

/* Add an active point into the model */
void _glassModelAddActivePoint(GlassModel *model, GlassActivePoint *apoint)
{
  if(!model || !apoint)
    return;
 
  /* Append */
  if(model->last_apoint)
  {
    model->last_apoint->next = apoint;
    model->last_apoint = apoint;
  }
  else
  {
    model->apoints = apoint;
    model->last_apoint = apoint;
  }

  /* Update the shortcut array */
  _glassModelBuildActivePointArray(model);
}

/* Find a active point with the name `name', return NULL if not found */
GlassActivePoint *_glassModelFindActivePoint(GlassModel *model, const char *name)
{
  GlassActivePoint *apoint_ptr;

  if(!model || !name)
    return NULL;

  apoint_ptr = model->apoints;
  while(apoint_ptr)
  {
    if(!strcmp(_glassActivePointGetName(apoint_ptr), name))
      return apoint_ptr;

    apoint_ptr = apoint_ptr->next;
  }

  /* Not found, so look in the shared library */
  if(model->library)
    return _glassLibraryFindActivePoint(model->library, name);

  return NULL;
}

/* Add a child node to the model */
void _glassModelAddChild(GlassModel *model, GlassNode *node)
{
  if(!model || !node)
    return;

  /* Append */
  if(model->last_child)
  {
    model->last_child->next = node;
    model->last_child = node;
  }
  else
  {
    model->children = node;
    model->last_child = node;
  }
}

/* (Re)Build both the variable and surface arrays */
void _glassModelBuildArrays(GlassModel *model)
{
  if(!model)
    return;

  _glassModelBuildSurfaceArray(model);
  _glassModelBuildVariableArray(model);
  _glassModelBuildActivePointArray(model);
}

/* Build an array of pointers to variables, for quick access */
static void _glassModelBuildVariableArray(GlassModel *model)
{
  GlassVariable *var_ptr;
  int i, nvariables = 0;

  if(!model)
    return;

  /* Count the number of variables */
  var_ptr = model->variables;
  while(var_ptr)
  {
    nvariables++;
    var_ptr = var_ptr->next;
  }

  /* Resize the array */
  if(nvariables == 0)
  {
    free(model->variable_array);
    model->variable_array = NULL;
  }
  else
    model->variable_array = (GlassVariable **)realloc(model->variable_array,
						       nvariables * sizeof(GlassVariable *));

  /* If no variables, set array to NULL */
  if(!nvariables)
    model->variable_array = NULL;

  /* Store pointers to the variables in order */
  var_ptr = model->variables;
  for(i = 0; i < nvariables; i++)
  {
    model->variable_array[i] = var_ptr;
    var_ptr = var_ptr->next;
  }

  /* Remeber how many variables there are */
  model->nvariables = nvariables;
}

/* Build an array of pointers to surfaces, for quick access */
static void _glassModelBuildSurfaceArray(GlassModel *model)
{
  GlassSurface *surface_ptr;
  int i, nsurfaces = 0;

  if(!model)
    return;

  /* Count the number of surfaces */
  surface_ptr = model->surfaces;
  while(surface_ptr)
  {
    nsurfaces++;
    surface_ptr = surface_ptr->next;
  }

  /* Resize the array */
  if(nsurfaces == 0)
  {
    free(model->surface_array);
    model->surface_array = NULL;
  }
  else
    model->surface_array = (GlassSurface **)realloc(model->surface_array,
						    nsurfaces * sizeof(GlassSurface *));

  /* If no surfaces, set array to NULL */
  if(!nsurfaces)
    model->surface_array = NULL;

  /* Store pointers to the surfaces in order */
  surface_ptr = model->surfaces;
  for(i = 0; i < nsurfaces; i++)
  {
    model->surface_array[i] = surface_ptr;
    surface_ptr = surface_ptr->next;
  }

  /* Remeber how many surfaces there are */
  model->nsurfaces = nsurfaces;
}

/* Build an array of pointers to active points, for quick access */
static void _glassModelBuildActivePointArray(GlassModel *model)
{
  GlassActivePoint *apoint_ptr;
  int i, npoints = 0;

  if(!model)
    return;

  /* Count the number of active points */
  apoint_ptr = model->apoints;
  while(apoint_ptr)
  {
    npoints++;
    apoint_ptr = apoint_ptr->next;
  }

  /* Resize the array */
  if(npoints == 0)
  {
    free(model->apoint_array);
    model->apoint_array = NULL;
  }
  else
    model->apoint_array = (GlassActivePoint **)realloc(model->apoint_array,
						       npoints * sizeof(GlassActivePoint *));

  /* If no active points, set array to NULL */
  if(!npoints)
    model->apoint_array = NULL;

  /* Store pointers to the active points in order */
  apoint_ptr = model->apoints;
  for(i = 0; i < npoints; i++)
  {
    model->apoint_array[i] = apoint_ptr;
    apoint_ptr = apoint_ptr->next;
  }

  /* Remeber how active points there are */
  model->npoints = npoints;
}

/* Go through the model and update the active points by doing the transforms */
void _glassModelUpdateActivePoints(GlassModel *model)
{
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();

  glLoadIdentity();

  _glassNodeUpdateActivePointsr(model->children);

  glPopMatrix();
}

/* Convert as many surfaces as possible into openGL display lists */
void _glassModelMakeDisplayLists(GlassModel *model)
{
  int i;

  if(!model)
    return;

  for(i = 0; i < model->nsurfaces; i++)
    _glassSurfaceMakeDisplayLists(model->surface_array[i]);
}

/* Free an model */
void _glassFreeModel(GlassModel *model)
{
  if(!model)
    return;

  if(model->name)
    free(model->name);

  free(model->surface_array);
  free(model->variable_array);
  free(model->apoint_array);

  free(model);
}
